# Opera Mail cipher selector
# by Matej Horvat
# February 23, 2020
# http://matejhorvat.si/en/opera/omail/

import os
import sys

from Tkinter import *
import tkFileDialog
import tkMessageBox

# Ciphers, sorted approximately from best to worst regarding forward secrecy,
# algorithm security, and key size
kCiphers = [
	0xC02B, "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256",
	0xC02F, "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256",
	0x009E, "TLS_DHE_RSA_WITH_AES_128_GCM_SHA256",
	0xC00A, "TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA",
	0xC014, "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA",
	0x0039, "TLS_DHE_RSA_WITH_AES_256_CBC_SHA",
	0xC009, "TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA",
	0xC013, "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA",
	0x0033, "TLS_DHE_RSA_WITH_AES_128_CBC_SHA",
	0x009C, "TLS_RSA_WITH_AES_128_GCM_SHA256",
	0x0035, "TLS_RSA_WITH_AES_256_CBC_SHA",
	0x002F, "TLS_RSA_WITH_AES_128_CBC_SHA",
	0x000A, "TLS_RSA_WITH_3DES_EDE_CBC_SHA",
	0xC007, "TLS_ECDHE_ECDSA_WITH_RC4_128_SHA",
	0xC011, "TLS_ECDHE_RSA_WITH_RC4_128_SHA",
	0x0005, "TLS_RSA_WITH_RC4_128_SHA",
	0x0004, "TLS_RSA_WITH_RC4_128_MD5",
]
kCipherMap = {}

def iterCiphers():
	i = 0
	while i < len(kCiphers):
		yield (kCiphers[i], kCiphers[i + 1])
		i += 2

def readByte(file):
	return ord(file.read(1))

def readWord(file):
	word = file.read(2)
	return (ord(word[0]) << 8) | ord(word[1])

def writeByte(file, value):
	file.write(chr(value))

def writeWord(file, value):
	file.write(chr((value >> 8) & 0xFF) + chr(value & 0xFF))

class App(Frame):
	def __init__(self, master):
		Frame.__init__(self, master)
		self.pack()

		self.master.title("Opera Mail cipher selector")

		# Create the buttons
		frame = Frame(self)
		frame.pack(fill=BOTH, expand=True)
		self.btnSave = Button(frame, text="Save", command=self.save)
		self.btnDefault = Button(
			frame, text="Set to default", command=self.setToDefault
		)
		self.btnRecommended = Button(
			frame, text="Set to recommended", command=self.setToRecommended
		)
		self.btnAbout = Button(frame, text="About", command=self.about)
		for btn in (
			self.btnSave, self.btnDefault, self.btnRecommended, self.btnAbout
		):
			btn.pack({"side": "left"})

		# Create a mapping of cipher ids to names
		for id, name in iterCiphers():
			kCipherMap[id] = name

		# Create the checkboxes and their variables (all off by default)
		self.variables = {id: IntVar(0) for id in kCipherMap.iterkeys()}
		self.chkCipher = \
			[
				Checkbutton(self, text=name, variable=self.variables[id]) \
				for id, name in iterCiphers()
			]
		for chk in self.chkCipher:
			chk.pack(anchor="w")

		# Open the opssl6.dat file
		try:
			self.opssl6 = open(
				os.environ["AppData"] + "\\Opera Mail\\Opera Mail\\opssl6.dat",
				"r+b"
			)
		except IOError:
			tkMessageBox.showinfo(
				"Locate opssl6.dat",
				"Please locate the \"opssl6.dat\" file. It is in your Opera Mail " + \
				"installation's \"profile\" folder."
			)
			filename = tkFileDialog.askopenfilename()
			if not filename:
				sys.exit(0)
			if not filename.lower().endswith("opssl6.dat"):
				tkMessageBox.showerror("Error", "Incorrect file chosen.")
				sys.exit(0)
			self.opssl6 = open(filename, "r+b")
		file = self.opssl6

		# Read the ids of enabled ciphers
		file.seek(0x20)
		numCiphers = readByte(file) / 2
		for i in xrange(numCiphers):
			self.variables[readWord(file)].set(1)

		# Read the rest (cached CA certificates?)
		self.theRest = file.read()

	def save(self):
		# Count the enabled ciphers
		numCiphers = 0
		for var in self.variables.itervalues():
			numCiphers += var.get()

		# Update the array lengths in the file
		file = self.opssl6
		file.seek(0x10)
		writeByte(file, numCiphers * 2 + 16)
		file.seek(0x1E)
		writeByte(file, numCiphers * 2 + 2)
		file.seek(0x20)
		writeByte(file, numCiphers * 2)

		# Write the ids of enabled ciphers; the order does not matter as Opera
		# Mail will always use its own preferred order
		for id, var in self.variables.iteritems():
			if var.get():
				writeWord(file, id)

		file.write(self.theRest)
		file.truncate()

		tkMessageBox.showinfo(
			"Saved",
			"If Opera Mail is running, please close it and click Save again.\n" + \
			"Then close this program and run Opera Mail."
		)

	def setAllTo(self, value):
		for var in self.variables.itervalues():
			var.set(value)

	def setToDefault(self):
		self.setAllTo(1)

	def setToRecommended(self):
		self.setAllTo(0)
		for id in (0xC02B, 0xC02F, 0x009E):
			self.variables[id].set(1)

	def about(self):
		tkMessageBox.showinfo(
			"About",
			"Opera Mail cipher selector\n" + \
			"by Matej Horvat\n" + \
			"\n" + \
			"February 23, 2020\n" + \
			"\n" + \
			"http://matejhorvat.si/en/opera/omail/\n" + \
			"\n" + \
			"This is NOT an Opera Software product!\n" +
			"Use at your own risk."
		)

root = Tk()
app = App(root)
app.mainloop()